
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_wasm.h>

#include <wasm_c_api.h>
#include <wasm_export.h>

#define ngx_wamr_err(b)  ngx_strnlen((u_char *) b, NGX_MAX_ERROR_STR - 1), b


typedef struct {
    size_t                 stack_size;
    size_t                 heap_size;
    size_t                 global_heap_size;
} ngx_wamr_conf_t;

typedef struct {
    wasm_module_inst_t     instance;
    wasm_exec_env_t        exec_env;
} ngx_wamr_instance_t;


static void  *ngx_wamr_create_conf(ngx_cycle_t *cycle);
static char *ngx_wamr_init_conf(ngx_cycle_t *cycle, void *conf);

static ngx_int_t ngx_wamr_create_runtime(ngx_cycle_t *cycle);
static void ngx_wamr_delete_runtime(ngx_cycle_t *cycle);

static ngx_int_t ngx_wamr_create_module(ngx_cycle_t *cycle,
    ngx_wasm_module_t *wm);
static void ngx_wamr_delete_module(ngx_wasm_module_t *wm);

static ngx_wasm_instance_t *ngx_wamr_create_instance(ngx_wasm_module_t *wm,
    ngx_pool_t *pool, ngx_log_t *log);
static ngx_int_t ngx_wamr_init_preopen(ngx_wasm_module_t *wm,
    ngx_pool_t *pool, ngx_log_t *log);
static void ngx_wamr_delete_instance(ngx_wasm_instance_t *wi);

static ngx_int_t ngx_wamr_enter(ngx_wasm_instance_t *wi,
    ngx_wasm_call_t *call);
void *ngx_wamr_translate(ngx_wasm_instance_t *wi, uint32_t addr, size_t len);


static int32_t ngx_wamr_open(wasm_exec_env_t exec_env, uint32_t namebuf,
    uint32_t len, uint32_t parent);
static int32_t ngx_wamr_close(wasm_exec_env_t exec_env, uint32_t handle);
static int32_t ngx_wamr_read(wasm_exec_env_t exec_env, uint32_t handle,
    uint32_t buf, uint32_t len);
static int32_t ngx_wamr_write(wasm_exec_env_t exec_env, uint32_t handle,
    uint32_t buf, uint32_t len);
static int32_t ngx_wamr_get(wasm_exec_env_t exec_env, uint32_t handle,
    uint32_t propid, uint32_t buf, uint32_t len);
static int32_t ngx_wamr_set(wasm_exec_env_t exec_env, uint32_t handle,
    uint32_t propid, uint32_t buf, uint32_t len);


static ngx_str_t  wamr_name = ngx_string("wamr");

static ngx_command_t  ngx_wamr_commands[] = {

    { ngx_string("wamr_stack_size"),
      NGX_WASM_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_size_slot,
      0,
      offsetof(ngx_wamr_conf_t, stack_size),
      NULL },

    { ngx_string("wamr_heap_size"),
      NGX_WASM_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_size_slot,
      0,
      offsetof(ngx_wamr_conf_t, heap_size),
      NULL },

    { ngx_string("wamr_global_heap_size"),
      NGX_WASM_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_size_slot,
      0,
      offsetof(ngx_wamr_conf_t, global_heap_size),
      NULL },

      ngx_null_command
};

static ngx_wasm_core_module_t  ngx_wamr_module_ctx = {
    &wamr_name,
    ngx_wamr_create_conf,              /* create configuration */
    ngx_wamr_init_conf,                /* init configuration */

    {
        ngx_wamr_create_runtime,
        ngx_wamr_delete_runtime,
        ngx_wamr_create_module,
        ngx_wamr_delete_module,
        ngx_wamr_create_instance,
        ngx_wamr_delete_instance,
        ngx_wamr_enter,
        ngx_wamr_translate,
        1,
    }
};


ngx_module_t  ngx_wamr_module = {
    NGX_MODULE_V1,
    &ngx_wamr_module_ctx,              /* module context */
    ngx_wamr_commands,                 /* module directives */
    NGX_WASM_MODULE,                    /* module type */
    NULL,                               /* init master */
    NULL,                               /* init module */
    NULL,                               /* init process */
    NULL,                               /* init thread */
    NULL,                               /* exit thread */
    NULL,                               /* exit process */
    NULL,                               /* exit master */
    NGX_MODULE_V1_PADDING
};


#define ngx_wasi_prefix  "ngx:wasi/syscall"

static NativeSymbol ngx_wamr_host_functions[] =
{
    { "open",  ngx_wamr_open,  "(iii)i",  NULL  },
    { "close", ngx_wamr_close, "(i)i",    NULL  },
    { "read",  ngx_wamr_read,  "(iii)i",  NULL  },
    { "write", ngx_wamr_write, "(iii)i",  NULL  },
    { "get",   ngx_wamr_get,   "(iiii)i", NULL  },
    { "set",   ngx_wamr_set,   "(iiii)i", NULL  },
};


static void *
ngx_wamr_create_conf(ngx_cycle_t *cycle)
{
    ngx_wamr_conf_t  *iwcf;

    iwcf = ngx_pcalloc(cycle->pool, sizeof(ngx_wamr_conf_t));
    if (iwcf == NULL) {
        return NULL;
    }

    iwcf->stack_size = NGX_CONF_UNSET_SIZE;
    iwcf->heap_size = NGX_CONF_UNSET_SIZE;
    iwcf->global_heap_size = NGX_CONF_UNSET_SIZE;

    return iwcf;
}


static char *
ngx_wamr_init_conf(ngx_cycle_t *cycle, void *conf)
{
    ngx_wamr_conf_t  *iwcf = conf;

    ngx_conf_init_size_value(iwcf->stack_size, 8092);
    ngx_conf_init_size_value(iwcf->heap_size, 8092);
    ngx_conf_init_size_value(iwcf->global_heap_size, (1024 * 1024));

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_wamr_create_runtime(ngx_cycle_t *cycle)
{
    u_char            *heap;
    RuntimeInitArgs    init_args;
    ngx_wamr_conf_t  *iwcf;

    iwcf = (ngx_wamr_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                  ngx_wamr_module);

    heap = ngx_pnalloc(cycle->pool, iwcf->global_heap_size);
    if (heap == NULL) {
        return NGX_ERROR;
    }

    ngx_memzero(&init_args, sizeof(RuntimeInitArgs));

    init_args.mem_alloc_type = Alloc_With_Pool;
    init_args.mem_alloc_option.pool.heap_buf = heap;
    init_args.mem_alloc_option.pool.heap_size = iwcf->global_heap_size;

    init_args.n_native_symbols = sizeof(ngx_wamr_host_functions)
                                 / sizeof(NativeSymbol);

    init_args.native_module_name = ngx_wasi_prefix;
    init_args.native_symbols = ngx_wamr_host_functions;

    if (!wasm_runtime_full_init(&init_args)) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wamr: wasm_runtime_init() failed");
        return NGX_ERROR;
    }

    return NGX_OK;
}


static void
ngx_wamr_delete_runtime(ngx_cycle_t *cycle)
{
    wasm_runtime_destroy();
}


static ngx_int_t
ngx_wamr_create_module(ngx_cycle_t *cycle, ngx_wasm_module_t *wm)
{
    uint8_t          *p;
    wasm_module_t     mod;
    ngx_wasm_conf_t  *wcf;

    char  err[NGX_MAX_ERROR_STR];

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                ngx_wasm_core_module);

    /* wamr modifies buffer on load, so duplicate to allow subsequent calls */
    p = ngx_palloc(cycle->pool, wm->src.len);
    if (p == NULL) {
        return NGX_ERROR;
    }

    memcpy(p, wm->src.data, wm->src.len);

    mod = wasm_runtime_load(p, wm->src.len, err, sizeof(err));
    if (mod == NULL) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wamr: wasm_runtime_load() failed: %*s",
                       ngx_wamr_err(err));
        return NGX_ERROR;
    }

    wm->data = mod;
    wm->ctx = ngx_wasm_get_conf(cycle->conf_ctx, ngx_wamr_module);
    wm->wcf = wcf;

    return NGX_OK;
}


static void
ngx_wamr_delete_module(ngx_wasm_module_t *wm)
{
    wasm_runtime_unload(wm->data);
}


static ngx_wasm_instance_t *
ngx_wamr_create_instance(ngx_wasm_module_t *wm, ngx_pool_t *pool,
    ngx_log_t *log)
{
    ngx_wamr_conf_t      *iwcf;
    wasm_module_inst_t     wmi;
    ngx_wasm_instance_t   *wi;
    ngx_wamr_instance_t  *iwi;

    char  error_buf[NGX_MAX_ERROR_STR];

    iwcf = (ngx_wamr_conf_t *) wm->ctx;

    wi = ngx_pcalloc(pool, sizeof(ngx_wasm_instance_t));
    if (wi == NULL) {
        return NULL;
    }

    iwi = ngx_pcalloc(pool, sizeof(ngx_wamr_instance_t));
    if (iwi == NULL) {
        return NULL;
    }

    wi->pool = pool;
    wi->log = log;
    wi->mod = wm;

    if (wm->preopen && wm->preopen->nelts) {
        ngx_wamr_init_preopen(wm, pool, log);
    }

    wmi = wasm_runtime_instantiate(wm->data, iwcf->stack_size, iwcf->heap_size,
                                   error_buf, sizeof(error_buf));
    if (wmi == NULL) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wamr: wasm_runtime_instantiate() failed: %*s",
                      ngx_wamr_err(error_buf));
        return NULL;
    }

    iwi->exec_env = wasm_runtime_create_exec_env(wmi, iwcf->stack_size);
    if (iwi->exec_env == NULL) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wamr: wasm_runtime_create_exec_env() failed");
        goto failed;
    }

    iwi->instance = wmi;

    wi->data = iwi;


    return wi;

failed:

    if (iwi->exec_env) {
        wasm_runtime_destroy_exec_env(iwi->exec_env);
    }

    wasm_runtime_deinstantiate(wmi);

    return NULL;
}

static ngx_int_t
ngx_wamr_init_preopen(ngx_wasm_module_t *wm, ngx_pool_t *pool, ngx_log_t *log)
{
    char               **map_dir_list, *p;
    size_t               size;
    ngx_uint_t           i;
    ngx_wasm_preopen_t  *po;

    po = wm->preopen->elts;

    size = 0;
    for (i = 0; i < wm->preopen->nelts; i++) {
        size += po[i].host.len + po[i].guest.len + sizeof("::");
    }

    map_dir_list = ngx_palloc(pool, wm->preopen->nelts * sizeof(char *) + size);
    if (map_dir_list == NULL) {
        return NGX_ERROR;
    }

    p = (char *) map_dir_list + wm->preopen->nelts * sizeof(char *);
    for (i = 0; i < wm->preopen->nelts; i++) {
        map_dir_list[i] = p;
        p = (char *) ngx_cpymem(p, po[i].guest.data, po[i].guest.len);
        *p++ = ':';
        *p++ = ':';
        p = (char *) ngx_cpymem(p, po[i].host.data, po[i].host.len);
        *p++ = 0;
        ngx_log_debug3(NGX_LOG_DEBUG_CORE, log, 0,
                       "module \"%V\" bind host:\"%V\" to guest:\"%V\"",
                       &wm->ns, &po[i].host, &po[i].guest);
    }

    wasm_runtime_set_wasi_args_ex(wm->data, NULL , 0,
                                  (const char **) map_dir_list,
                                  wm->preopen->nelts,
                                  NULL, 0, NULL, 0, -1, -1, -1);
    return NGX_OK;
}


static ngx_int_t
ngx_wamr_enter(ngx_wasm_instance_t *wi, ngx_wasm_call_t *call)
{
    ngx_wamr_instance_t *iwi = wi->data;

    const char            *ex;
    ngx_uint_t             i;
    wasm_val_t            *fargs, res;
    wasm_function_inst_t   func;

    func = wasm_runtime_lookup_function(iwi->instance,
                                        (char *) call->name.data, NULL);
    if (func == NULL) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wamr: wasm_runtime_lookup_function(%V) failed",
                      &call->name);
        return NGX_ERROR;
    }

    wasm_runtime_set_user_data(iwi->exec_env, call->host);

    switch (call->rc.kind) {
        case NGX_WT_I32:
            res.kind = WASM_I32;
            res.of.i32 = 0;
            break;
        case NGX_WT_I64:
            res.kind = WASM_I64;
            res.of.i64 = 0;
            break;
        case NGX_WT_F32:
            res.kind = WASM_F32;
            res.of.f32 = 0;
            break;
        case NGX_WT_F64:
            res.kind = WASM_F64;
            res.of.f64 = 0;
            break;

        default:
            ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "unknown wasm type");

            return NGX_ERROR;
    }

    if (call->nargs) {

        fargs = ngx_pcalloc(call->host->pool,
                            sizeof(wasm_val_t) * call->nargs);
        if (fargs == NULL) {
            return NGX_ERROR;
        }

        for (i = 0; i < call->nargs; i++) {
            switch (call->args[i].kind) {
            case NGX_WT_I32:
                fargs[i].kind = WASM_I32;
                fargs[i].of.i32 = call->args[i].of.i32;
                break;

            case NGX_WT_I64:
                fargs[i].kind = WASM_I64;
                fargs[i].of.i64 = call->args[i].of.i64;
                break;

            case NGX_WT_F32:
                fargs[i].kind = WASM_F32;
                fargs[i].of.f32 = call->args[i].of.f32;
                break;

            case NGX_WT_F64:
                fargs[i].kind = WASM_F64;
                fargs[i].of.f64 = call->args[i].of.f64;
                break;

            default:
                ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "unknown wasm type");

                return NGX_ERROR;
            }
        }

    } else {
        fargs = NULL;
    }

    ngx_log_debug3(NGX_LOG_DEBUG_CORE, wi->log, 0,
                   "wamr: calling %V.%V(%uD args)", &wi->mod->ns,
                   &call->name, call->nargs);

    if (!wasm_runtime_call_wasm_a(iwi->exec_env, func, 1, &res,
                                  call->nargs, fargs))
    {
        ex = wasm_runtime_get_exception(iwi->instance);

        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wamr: wasm_runtime_call_wasm(%V) failed: %s",
                      &call->name, ex ? ex : "" );

        return NGX_ERROR;
    }

    switch (res.kind) {
        case WASM_I32:
            call->rc.of.i32 = res.of.i32;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wamr: call \"%V\" rc:%D",
                           &call->name, call->rc.of.i32);
            break;

        case WASM_I64:
            call->rc.of.i64 = res.of.i64;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wamr: call \"%V\" rc:%L",
                           &call->name, call->rc.of.i64);
            break;

        case WASM_F32:
            call->rc.of.f64 = res.of.f32;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wamr: call \"%V\" rc:%f",
                           &call->name, call->rc.of.f32);
            break;

        case WASM_F64:
            call->rc.of.f64 = res.of.f64;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wamr: call \"%V\" rc:%f",
                           &call->name, call->rc.of.f64);
            break;

        default:
            ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "unknown wasm type");

            return NGX_ERROR;
    }

    return NGX_OK;
}


void *
ngx_wamr_translate(ngx_wasm_instance_t *wi, uint32_t addr, size_t len)
{
    ngx_wamr_instance_t *iwi = wi->data;

    if (!wasm_runtime_validate_app_addr(iwi->instance, addr, len)) {
        return NULL;
    }

    return wasm_runtime_addr_app_to_native(iwi->instance, addr);
}


static void
ngx_wamr_delete_instance(ngx_wasm_instance_t *wi)
{
    ngx_wamr_instance_t *iwi = wi->data;

    wasm_runtime_destroy_exec_env(iwi->exec_env);
    wasm_runtime_deinstantiate(iwi->instance);
}


static int32_t
ngx_wamr_open(wasm_exec_env_t exec_env, uint32_t namebuf, uint32_t len,
    uint32_t parent)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    uint8_t  *buf;

    buf = ngx_wasm_translate(host->wi, namebuf, len);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wamr_read bad address");
        return NGX_ERROR;
    }

    return host->ops->open(host, buf, len, parent);
}


static int32_t
ngx_wamr_close(wasm_exec_env_t exec_env, uint32_t handle)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    return host->ops->close(host, handle);
}


static int32_t
ngx_wamr_read(wasm_exec_env_t exec_env, uint32_t handle, uint32_t b,
    uint32_t len)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    uint8_t  *buf;

    buf = ngx_wasm_translate(host->wi, b, len);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wamr_read bad address");
        return NGX_ERROR;
    }

    return host->ops->read(host, handle, buf, len);
}


static int32_t
ngx_wamr_write(wasm_exec_env_t exec_env, uint32_t handle, uint32_t b,
    uint32_t len)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    uint8_t  *buf;

    buf = ngx_wasm_translate(host->wi, b, len);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wamr_write bad address");
        return NGX_ERROR;
    }

    return host->ops->write(host, handle, buf, len);
}


static int32_t
ngx_wamr_get(wasm_exec_env_t exec_env, uint32_t handle, uint32_t propid,
    uint32_t b, uint32_t len)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    uint8_t  *buf;

    buf = ngx_wasm_translate(host->wi, b, len);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wamr_get bad address");
        return NGX_ERROR;
    }

    return host->ops->get(host, handle, propid, buf, len);
}


static int32_t
ngx_wamr_set(wasm_exec_env_t exec_env, uint32_t handle, uint32_t propid,
    uint32_t b, uint32_t len)
{
    ngx_wasm_host_t *host = wasm_runtime_get_user_data(exec_env);

    uint8_t  *buf;

    buf = ngx_wasm_translate(host->wi, b, len);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wamr_set bad address");
        return NGX_ERROR;
    }

    return host->ops->set(host, handle, propid, buf, len);
}
