
/*
 * Copyright (C) 2024 Web Server LLC
 */


#ifndef _NGX_WASM_HOST_H_INCLUDED_
#define _NGX_WASM_HOST_H_INCLUDED_


#include <ngx_config.h>
#include <ngx_core.h>

#define ngx_wasm_host_null_ops                                                \
    { ngx_null_string, 0, NULL, NULL, NULL, NULL, NULL, NULL }


typedef struct ngx_wasm_host_s            ngx_wasm_host_t;
typedef struct ngx_wasm_handle_s          ngx_wasm_handle_t;
typedef struct ngx_wasm_handles_s         ngx_wasm_handles_t;
typedef struct ngx_wasm_handle_ops_s      ngx_wasm_handle_ops_t;

#include <ngx_wasm_handle.h>


typedef struct {
    int32_t (*open)(ngx_wasm_host_t *host, uint8_t *namebuf, uint32_t len,
                    int32_t ctx);
    int32_t (*close)(ngx_wasm_host_t *host, int32_t handle);
    int32_t (*read)(ngx_wasm_host_t *host, int32_t handle, uint8_t *buf,
                    uint32_t len);
    int32_t (*write)(ngx_wasm_host_t *host, int32_t handle, uint8_t *buf,
                     uint32_t len);
    int32_t (*get)(ngx_wasm_host_t *host, int32_t handle, uint32_t propid,
                   uint8_t *buf, uint32_t len);
    int32_t (*set)(ngx_wasm_host_t *host, int32_t handle, uint32_t propid,
                   uint8_t *buf, uint32_t len);
} ngx_wasm_host_ops_t;


/* the outermost context for wasm calls */
struct  ngx_wasm_host_s {
    ngx_wasm_instance_t        *wi;
    ngx_pool_t                 *pool;
    ngx_log_t                  *log;
    ngx_str_t                   name;            /* host name for logs */
    ngx_wasm_host_ops_t        *ops;             /* wasm engine entries */
    ngx_wasm_handles_t         *handles;         /* open descriptors */
    ngx_hash_t                  handle_ops;      /* per-object operations */
    ngx_hash_keys_arrays_t      handle_ops_keys;
    ngx_queue_t                 events;          /* ngx events to process */
    ngx_queue_t                 free_events;     /* reported evs return here */
    ngx_event_t                *fd_event;
    /* extended by specific host data on creation */
};


typedef struct  {
    ngx_queue_t                  queue;
    int32_t                      fd;
    uint32_t                     ptr;
    uint32_t                     flags;
} ngx_wasm_event_t;


ngx_wasm_host_t *ngx_wasm_host_create(ngx_wasm_instance_t *wi, ngx_pool_t *pool,
    size_t size, ngx_log_t *log, const char *name, ngx_array_t *apis,
    ngx_array_t *acl);

/*
 * creates parent handle with host context for open() of specified API;
 * attempt to use such handle for actual operations will lead to using
 * stub ops; thus it is only used to pass context to wasm as descriptor
 */
int32_t ngx_wasm_host_create_api_context(ngx_wasm_host_t *host,
    const char *api, void *data);
void *ngx_wasm_host_get_api_context(ngx_wasm_handle_t *handle,
    ngx_wasm_handle_t *ph);
void *ngx_wasm_host_get_object_context(ngx_wasm_handle_t *handle,
    ngx_wasm_handle_t *ph, ngx_uint_t tag);

/* create context AND perform real open using the context */
int32_t ngx_wasm_host_open_object(ngx_wasm_host_t *wh, const char *api,
    void *data);
/* open given API in the context of given handle */
ngx_int_t ngx_wasm_host_open_as(ngx_wasm_host_t *wh, ngx_wasm_handle_t *handle,
    const char *api, void *data);

/* create object without calling open() method - for simple objs without it */
int32_t ngx_wasm_host_create_object(ngx_wasm_host_t *host, const char *api,
    void *data);
ngx_int_t ngx_wasm_host_delete_object(ngx_wasm_host_t *host, int32_t fd);

ngx_wasm_event_t *ngx_wasm_host_alloc_event(ngx_wasm_host_t *host);

int32_t ngx_wasm_host_open_stub(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *parent);
int32_t ngx_wasm_host_io_stub(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);
int32_t ngx_wasm_host_property_stub(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
int32_t ngx_wasm_host_close_stub(ngx_wasm_handle_t *self);


#endif /* _NGX_WASM_HOST_H_INCLUDED_ */
